/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#ifndef MT_control_h
#define MT_control_h

#include "array.h"
#include "ors.h"

struct ControlVariable;

typedef MT::Array<ControlVariable*> ControlVariableList;

extern ControlVariableList globalSpace;


//===========================================================================
//
// Control Variable definition
//

/*!\brief different types of control variables: refer to different ways to
  compute/access the kinematics and Jacobians */
enum CVtype {
  noneCVT,     //!< undefined
  posCVT,      //!< 3D position of reference, can have 2nd reference, no param
  zoriCVT,     //!< 3D z-axis orientation, no 2nd reference, no param
  zalignCVT,   //!< 1D z-axis alignment, can have 2nd reference, no param
  qLinearCVT,  //!< k-dim variable linear in q, no references, param: k-times-n matrix
  qSingleCVT,  //!< 1D entry of q, reference-integer=index, no param
  qSquaredCVT, //!< 1D square norm of q, no references, param: n-times-n matrix
  qLimitsCVT,  //!< 1D meassure for joint limit violation, no references, param: n-times-2 matrix with lower and upper limits for each joint
  collCVT,     //!< 1D meassure for collision violation, no references, param: 1D number defining the distance margin
  comCVT,      //!< 2D vector of the horizontal center of mass, no refs, no param
  skinCVT,     //!< vector of skin pressures...
  gripCVT, rotCVT, contactCVT //PRELIMINARY OR OBSOLETE
};

enum TargetType{ noneTT, directTT, gainsTT, attractorTT, trajectoryTT };

/*!\brief basic control variable */
struct ControlVariable{
  //!@name data fields
  bool active;          //!< active?
  CVtype type;          //!< which type has this variable
  TargetType targetType;//!< what target type
  MT::String name;      //!< its name
  ors::Graph *ors;      //!< pointer to the data structure (from which it gets the kinematics)
  int i,j;              //!< which body(-ies) does it refer to?
  ors::Frame irel,jrel; //!< relative position to the body
  arr params;           //!< parameters of the variable (e.g., liner coefficients, limits, etc)

  arr x,x_old,v,v_old,x_target,v_target;     //!< current state and final target of this variable
  arr J,tJ;                                  //!< current Jacobian and its transpose
  double prec,vprec;                         //!< precision (=1/variance) associated with this variable
  arr x_trajectory,prec_trajectory;          //!< target & precision over a whole trajectory
  arr v_trajectory,vprec_trajectory;         //!< target & precision over a whole trajectory

  arr x_change;                              //!< desired change for online targets
  double a,b;                                //!< parameters of the PD controller or attractor dynamics
  double err,derr;
  int state;                                 //!< discrete indicate state of this variable (e.g., convergence)
  double state_tol;

  //!@name initialization
  ControlVariable();
  ControlVariable(
    const char* _name,
    ors::Graph& _sl,
    CVtype _type,
    const char *iname,const char *iframe,
    const char *jname,const char *jframe,
    const arr& _params);
  ~ControlVariable();

  void set(
    const char* _name,
    ors::Graph &_sl,
    CVtype _type,
    int _i,const ors::Frame& _irel,
    int _j,const ors::Frame& _jrel,
    const arr& _params);
  //void set(const char* _name,ors::Graph& _sl,CVtype _type,const char *iname,const char *jname,const char *reltext);

  //!@name online target parameters
  void setGains(double _a,double _b);
  void setGainsAsAttractor(double decaySteps,double oscillations=.2);
  
  //!@name trajectory target parameters
  void setConstantTargetTrajectory(uint T);
  void setInterpolatedTargetTrajectory(uint T);
  void setPrecisionTrajectoryFinal(uint T,double intermediate_prec,double final_prec);
  void setPrecisionTrajectoryConstant(uint T,double constant_prec);
  void setPrecisionVTrajectoryFinal(uint T,double intermediate_prec,double final_prec);
  void setPrecisionVTrajectoryConstant(uint T,double constant_prec);
  void setTrajectory(uint T,double funnelsdv=0.,double funnelvsdv=0.); //OBSOLETE
  
  void setInterpolatedTargetsEndPrecisions(uint T,double inter_yprec,double end_yprec,double inter_vprec,double end_vprec);
  void setInterpolatedTargetsConstPrecisions(uint T,double yprec,double vprec);

  //!@name updates
  void updateState();
  void updateJacobian();
  void updateChange(int t=-1);
  void getHessian(arr& H);

  //!@name I/O
  void write(ostream& os) const;
};

inline std::ostream& operator<<(std::ostream& os,const ControlVariable& x){ x.write(os); return os; }


//===========================================================================
//
// ControlVariableList functions
//

void reportAll   (ControlVariableList& CS,ostream& os,bool onlyActives=true);
void reportState (ControlVariableList& CS,ostream& os,bool onlyActives=true);
void reportErrors(ControlVariableList& CS,ostream& os,bool onlyActives=true,int t=-1);
void reportNames (ControlVariableList& CS,ostream& os,bool onlyActives=true);

void activateAll(ControlVariableList& CS,bool active);

void updateState(ControlVariableList& CS);
void updateJacobian(ControlVariableList& CS);
void updateChanges(ControlVariableList& CS,int t=-1);

void getJointJacobian(ControlVariableList& CS,arr& J);
void getJointXchange(ControlVariableList& CS,arr& x_change);


//***** OUTDATED CODE, replaced by equivalents in the SocAbstraction
double getLogLikelihood_obsolete(ControlVariableList& CS,int t);
double getCost_obsolete(ControlVariableList& CS,const arr& W,int t);
void getCostGradient_obsolete(ControlVariableList& CS,arr& dCdq,const arr& W,int t);
void hierarchicalControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W);
void naiveHierarchicalControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W);
void bayesianControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W);
void additiveControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W);
void bayesianPlanner_obsolete(ors::Graph *ors,ControlVariableList& CS,
                     arr& q,uint T,const arr& W,uint iterations,
                     std::ostream* os,int display,bool repeat);
//void SMAC::readCVdef(std::istream& is){
//void CVclear();
//void plotCVs();


#ifdef MT_IMPLEMENTATION
#  include"ors_control.cpp"
#endif

#endif
