/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#ifndef MT_ors_ode_h
#define MT_ors_ode_h

#include "ors.h"

struct dxWorld;		/* dynamics world */
struct dxSpace;		/* collision space */
struct dxBody;		/* rigid body (dynamics object) */
struct dxGeom;		/* geometry (collision object) */
struct dxJoint;
struct dxJointNode;
struct dxJointGroup;
struct dContactGeom;

typedef struct dxWorld *dWorldID;
typedef struct dxSpace *dSpaceID;
typedef struct dxBody *dBodyID;
typedef struct dxGeom *dGeomID;
typedef struct dxJoint *dJointID;
typedef struct dxJointGroup *dJointGroupID;

//===========================================================================
//
// Ode interface
//

/*! A trivial interface to the Open Dynamic Engine library. \ingroup sl

  It basically contains a dSpace and dWorld, provides a proximity
  callback function, and basic stepping function */
class OdeModule{
public:
  double time;
  dxSpace *space;
  dxGeom *plane0,*planex1,*planex2,*planey1,*planey2;
  dxWorld *world;
  dxJointGroup *contactgroup;
  double ERP,CFM; //integration parameters
  double coll_ERP,coll_CFM,coll_bounce,friction; //collision parameter
  bool noGravity,noContactJoints;

  MT::Array<dxBody*> bodies;
  MT::Array<dxJoint*> joints;
  MT::Array<dxJoint*> motors;
  MT::Array<dContactGeom*> conts;

public:
  OdeModule();
  ~OdeModule();

  /*!\brief reinstantiates a new ODE world (and space) clear of all previous objects */
  void clear();

  /*!\brief this function is called from the `dSpaceCollide' routine (in the
     `step' routine) when two objects get too close. A
     "collision-joint" is inserted between them that exerts the force
     of the collision. All of these collision-joints are collected in
     a group, and they are deleted after the `dWorldStep' by the
     `dJoinGroupEmpty' routine (in the `step' routine) */
  static void staticCallback(void *classP, dxGeom *g1, dxGeom *g2);

  //! sets gravity to zero (or back to -9.81)
  void setForceFree(bool free);

  /*!\brief main method: process one time step by calling SpaceCollide and WorldQuickStep */
  void step(double dtime=.01);

  void printInfo(std::ostream& os,dxBody *b);
  void reportContacts();
  void contactForces();
  void penetration(ors::Vector &p);
};


//===========================================================================
//
// communication with Ode
//

/*!\brief copy all frame variables (positions, velocities, etc)
into the ODE engine (createOde had to be called before) */
void exportStateToOde(ors::Graph &C,OdeModule& ode);

/*!\brief copy the current state of the ODE engine into
the frame variables (positions, velocities, etc) */
void importStateFromOde(ors::Graph &C,OdeModule& ode);

/*!\brief copy all frame variables (positions, velocities, etc)
into the ODE engine (createOde had to be called before) */
void exportForcesToOde(ors::Graph &C,OdeModule& ode);

void setJointForce(ors::Graph &C,ors::Graph::edge e,double f1,double f2);

/*!\brief applies forces on the configuration as given by the force vector x */
void setJointForce(ors::Graph &C,arr& f);

/*!\brief returns the number of motors attached to joints */
uint getJointMotorDimension(ors::Graph &C);

/*!\brief set the desired positions of all motor joints */
void setJointMotorPos(ors::Graph &C,OdeModule& ode,arr& x,double maxF=1.,double tau=.01);
/*!\brief set the desired position of a specific motor joint */
void setJointMotorPos(ors::Graph &C,OdeModule& ode,ors::Graph::edge e,double x0,double maxF=1.,double tau=.01);

/*!\brief set the desired velocities of all motor joints */
void setJointMotorVel(ors::Graph &C,OdeModule& ode,arr& v,double maxF=1.);
/*!\brief set the desired velocity of a specific motor joint */
void setJointMotorVel(ors::Graph &C,OdeModule& ode,ors::Graph::edge e,double v0,double maxF=1.);

/*!\brief disable motors by setting maxForce parameter to zero */
void unsetJointMotors(ors::Graph &C,OdeModule& ode);
/*!\brief disable motor by setting maxForce parameter to zero */
void unsetJointMotor(ors::Graph &C,OdeModule& ode,ors::Graph::edge e);

/*!\brief get the forces induced by all motor motor joints */
void getJointMotorForce(ors::Graph &C,OdeModule& ode,arr& f);
/*!\brief get the force induced by a specific motor joint */
void getJointMotorForce(ors::Graph &C,OdeModule& ode,ors::Graph::edge e,double& f);

void pidJointPos(ors::Graph &C,OdeModule& ode,ors::Graph::edge e,double x0,double v0,double xGain,double vGain,double iGain=0,double* eInt=0);

void pidJointVel(ors::Graph &C,OdeModule& ode,ors::Graph::edge e,double v0,double vGain);

//! [obsolete] \ingroup sl
void getGroundContact(ors::Graph &C,OdeModule& ode,boolA& cts);

//! import the information from ODE's contact list into the proximity list \ingroup sl
void importProxiesFromOde(ors::Graph &C,OdeModule& ode);

/*!\brief simulates one time step with the ODE engine, starting from state
    vector in, returns state vector out \ingroup sl */
void step(ors::Graph &C,OdeModule& ode,arr& in,arr& force,arr& out,uint steps=1);

/*!\brief simulates one time step with the ODE engine, starting from state
    vector in, returns state vector out \ingroup sl */
void step(ors::Graph &C,OdeModule& ode,arr& force,arr& out,uint steps=1);

//! \ingroup sl
void step(ors::Graph &C,OdeModule& ode,uint steps=1,double tau=.01);

/*!\brief instantiate the configuration in an ODE engine (first clears the
    ODE engine from all other objects) \ingroup sl */
void createOde(ors::Graph &C,OdeModule& ode);

//! \ingroup sl
void slGetProxies(ors::Graph &C,OdeModule &ode);

//! \ingroup sl
//void slGetProxyGradient(arr &dx,const arr &x,ors::Graph &C,OdeModule &ode);

struct qGrad{
  ors::Graph *C;
  OdeModule *ode;
  //OpenGL *gl;
  uint T,n,act; ors::Frame rel; arr q0,goal;
  double costCollision,costGoal;
  arr costW;
  qGrad(){ costCollision=10.; costGoal=50.; }

  void init(ors::Graph *_CC,OdeModule *_ode,arr &_q,arr &_goal,uint _act,ors::Frame &_rel,arr &_W){
    C=_CC; ode=_ode; //gl=_gl;
    CHECK(_q.nd==2,"");
    q0=_q[0]; goal=_goal; act=_act; rel=_rel;
    T=_q.d0;
    n=_q.d1;
    costW=_W;
  }
  double map(const arr &q){
    ((arr&)q).reshape(T,n);
    arr z(T,3);
    uint t,i;

    q[0]()=q0;
    //foward pass
    double cost=0.;
    for(t=0;t<T;t++){
      C->setJointState(q[t]);
      C->calcNodeFramesFromEdges();
      C->kinematics(z[t](),act,&rel);
      if(t==T-1) cost += costGoal*sqrDistance(goal ,z[t]);
      if(t>0)    cost += sqrDistance(costW,q[t-1],q[t]);
      slGetProxies(*C,*ode);
      for(i=0;i<C->proxies.N;i++) cost += costCollision*MT::sqr(C->proxies(i)->d);
    }
#ifdef MT_plot_h
    plotClear();
    plotLine(z);
#endif
    return cost;
  }
  void grad(arr &dq,const arr &q){
    ((arr&)q).reshape(T,n);
    dq.resize(T,n);
    arr z(3);
    uint t;

    arr J_T,z_T,cGrad;
    
    q[0]()=q0;
    //final state
    C->setJointState(q[0]);
    C->calcNodeFramesFromEdges();
    C->kinematics(z_T,act,&rel);
    C->jacobian  (J_T,act,&rel);
    //backward pass
    for(t=T;t--;){
      dq[t] = 0.;
      if(t==T-1) dq[t]() += 2.* costGoal * ~(J_T) * (z_T-goal);
      if(t<T-1)  dq[t]() += 2.* costW * (q[t]-q[t+1]);
      if(t>0)    dq[t]() += 2.* costW * (q[t]-q[t-1]);

      //slGetProxyGradient(cGrad,q[t],*C,*ode);
      C->setJointState(q[t]);
      C->calcNodeFramesFromEdges();
      slGetProxies(*C,*ode);
      C->getContactGradient(cGrad);
      dq[t]() -= costCollision * cGrad;
      //gl->display();
    }
    dq[0]()=0.;
    dq.reshape(dq.N);
  }
};

//! \ingroup sl
void reportContacts(OdeModule& ode);
//! \ingroup sl
bool inFloorContacts(OdeModule& ode,ors::Vector& x);


#ifdef MT_IMPLEMENTATION
#  include "ors_ode.cpp"
#endif

#endif
